/*!
 * \file bitMapMarker.cpp
 * \brief This file is used to store and control of bit map.
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

#include <sys/mman.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <stddef.h>

#include "bitMapMarker.hpp"
#include "util.hpp"

#ifdef __LP64__
    
    /*!
     * \brief Memory align bit count.
     *        Pointer is aligned 8byte,
     *        If allocate pointer as like "void*" in 64bit.
     *        So, 3bits under allocated pointer is always 0.
     */
    #define MEMALIGN_BIT 3
    
    /*!
     * \brief Bit size of pointer.
     *        E.g. "sizeof(void*) << 3".
     */
    #define PTR_BIT_WIDTH 64
    
    /*!
     * \brief Log. 2 of allocated pointer size.
     *        E.g. "log 2 sizeof(void*)".
     */
    #define LOG2_PTR_WIDTH 6
    
#else // __ILP32__
    
    /*!
     * \brief Memory align bit count.
     *        Pointer is aligned 4byte,
     *        If allocate pointer as like "void*" in 32bit.
     *        So, 2bits under allocated pointer is always 0.
     */
    #define MEMALIGN_BIT 2
    
    /*!
     * \brief Bit size of pointer.
     *        E.g. "sizeof(void*) << 3".
     */
    #define PTR_BIT_WIDTH 32
    
    /*!
     * \brief Log. 2 of allocated pointer size.
     *        E.g. "log 2 sizeof(void*)".
     */
    #define LOG2_PTR_WIDTH 5
    
#endif

/*!
 * \brief TBitMapMarker constructor.
 * \param startAddr [in] Start address of Java heap.
 * \param size      [in] Max Java heap size.
 */
TBitMapMarker::TBitMapMarker(void *startAddr, size_t size) {
    
    /* Sanity check. */
    if (unlikely(startAddr == NULL || size <= 0)) {
        throw -1;
    }
    
    /* Initialize setting. */
    size_t alignedSize = ALIGN_SIZE_UP(size, systemPageSize);
    
    this->beginAddr  = startAddr;
    this->endAddr    = incAddress(this->beginAddr, alignedSize);
    this->bitmapSize = ALIGN_SIZE_UP(alignedSize >> MEMALIGN_BIT,
        systemPageSize);
    this->bitmapAddr = mmap(NULL, this->bitmapSize, PROT_READ | PROT_WRITE,
        MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
    
    /* If failure allocate bitmap memory. */
    if (unlikely(this->bitmapAddr == MAP_FAILED)) {
        throw errno;
    }
    
    /* Advise the kernel that this memory will be random access. */
    madvise(this->bitmapAddr, this->bitmapSize, POSIX_MADV_RANDOM);
    
    /* Initialize bitmap (clear bitmap). */
    this->clear();
}

/*!
 * \brief TBitMapMarker destructor.
 */
TBitMapMarker::~TBitMapMarker() {
    
    /* Release memory map. */
    munmap(this->bitmapAddr, this->bitmapSize);
}

/*!
 * \brief Get a byte data and mask expressing target pointer.
 * \param addr  [in]  Oop address.
 * \param block [out] A byte data including bit flag.
 *                    The bit flag is expressing target pointer.
 * \param mask  [out] A bitmask for getting only flag of target pointer.
 */
inline void TBitMapMarker::getBlockAndMask(void *addr, ptrdiff_t **block,
    ptrdiff_t *mask) {
    
    /*
     * Calculate bitmap offset.
     * offset = (address offset from top of JavaHeap) / (pointer size)
     */
    ptrdiff_t bitmapOfs = ((ptrdiff_t)addr - (ptrdiff_t)this->beginAddr)
        >> MEMALIGN_BIT;
    
    /*
     * Calculate bitmap block.
     * block = (top of bitmap) + ((bitmap offset) / (pointer size(bit)))
     */
    *block = (ptrdiff_t *)this->bitmapAddr + (bitmapOfs >> LOG2_PTR_WIDTH);
    
    /*
     * Calculate bitmask.
     * mask = 1 << ((bitmap offset) % (pointer size(bit)))
     */
    *mask = (ptrdiff_t)1 << (bitmapOfs % PTR_BIT_WIDTH);
}

/*!
 * \brief Mark GC-marked address in this bitmap.
 * \param addr [in] Oop address.
 */
void TBitMapMarker::setMark(void *addr) {
    
    /* Sanity check. */
    if (unlikely(!this->isInZone(addr))) {
        return;
    }
    
    /*
     * Why I use "ptrdiff_t" ?
     *  ptrdiff_t is integer representation of pointer. ptrdiff_t size equals
     *  pointer size. Pointer size equals CPU register size.
     *  Thus, we use CPU register more effective.
     */
    ptrdiff_t *bitmapBlock;
    ptrdiff_t bitmapMask;
    this->getBlockAndMask(addr, &bitmapBlock, &bitmapMask);
    
    /* Atomic set mark bit flag. */
    asm volatile(
        #ifdef __LP64__
            "lock orq %1, %0;"
        #else // __ILP32__
            "lock orl %1, %0;"
        #endif
        : "+m" (*bitmapBlock)
        : "r" (bitmapMask)
        : "cc", "memory");
}

/*!
 * \brief Get marked flag of designated pointer.
 * \param addr [in] Targer pointer.
 * \return Designated pointer is marked.
 */
bool TBitMapMarker::isMarked(void *addr) {
    
    /* Sanity check. */
    if (unlikely(!this->isInZone(addr))) {
        return false;
    }
    
    /* Get block and mask for getting mark flag. */
    ptrdiff_t *bitmapBlock;
    ptrdiff_t bitmapMask;
    this->getBlockAndMask(addr, &bitmapBlock, &bitmapMask);
    
    /* Atomic get mark bit flag. */
    register bool result asm("al");
    /*
     * Intel 64 and IA-32 Architectures Software Developer's Manual.
     * Volume 3A: System Programming Guide, Part1
     *  8.2.3.8 Locked Instructions Have a Total Order:
     *    The memory-ordering model ensures that all processors agree
     *    on a single execution order of all locked instructions,
     *    including those that are larger than 8 bytes
     *    or are not naturally aligned.
     */
    asm volatile(
        "test %1, %2;"
        "setneb %0;"
        : "=r" (result)
        : "m" (*bitmapBlock), "r" (bitmapMask)
        : "cc", "memory");
    
    return result;
}

/*!
 * \brief Check address which is already marked and set mark.
 * \param addr [in] Oop address.
 * \return Designated pointer is marked.
 */
bool TBitMapMarker::checkAndMark(void *addr) {
    
    /* Sanity check. */
    if (unlikely(!this->isInZone(addr))) {
        return false;
    }
    
    /* Get block and mask for getting mark flag. */
    ptrdiff_t *bitmapBlock;
    ptrdiff_t bitmapMask;
    this->getBlockAndMask(addr, &bitmapBlock, &bitmapMask);
    
    /* Get and set mark. */
    register bool result asm("al");
    asm volatile(
        #ifdef __amd64__
            "bsfq %1, %%rbx;"
            "lock btsq %%rbx, %2;"
        #else // __i386__
            "bsfl %1, %%edx;"
            "lock btsl %%edx, %2;"
        #endif
            "setbb %0;"
        : "=r" (result)
        : "r" (bitmapMask), "m" (*bitmapBlock)
        :
        #ifdef __amd64__
            "%rbx",
        #else // __i386__
            "%edx",
        #endif
            "cc", "memory");
    
    return result;
}

/*!
 * \brief Clear bitmap flag.
 */
void TBitMapMarker::clear(void) {
    /*
     * Linux memory page size = 4KB.
     * So top address of _bitmap is 16byte aligned.
     * _size is ditto.
     *
     * memset() in glibc sets 128bytes per loop.
     *
     *******************************************************
     * Intel 64 and IA-32 Architectures Optimization Reference Manual
     *   Assembler/Compiler Coding Rule 12:
     *     All branch targets should be 16-byte aligned.
     */
    
    /* Temporary advise to zero-clear with sequential access. */
    madvise(this->bitmapAddr, this->bitmapSize, MADV_SEQUENTIAL);
    
    #ifdef AVX
        if (likely(usableAVX)) {
            asm volatile(
                "vxorpd %%ymm0, %%ymm0, %%ymm0;"
                ".align 16;"
                "AVX_LOOP:" /* memset 128bytes per LOOP. */
                    "vmovntdq %%ymm0, -0x80(%1, %0, 1);"
                    "vmovntdq %%ymm0, -0x60(%1, %0, 1);"
                    "vmovntdq %%ymm0, -0x40(%1, %0, 1);"
                    "vmovntdq %%ymm0, -0x20(%1, %0, 1);"
                #ifdef __amd64__
                    "subq $0x80, %0;"
                #else // __i386__
                    "subl $0x80, %0;"
                #endif
                "jnz AVX_LOOP;"
                :
                : "r" (this->bitmapSize), "r" (this->bitmapAddr)
                : "cc" /*, "%ymm0" */);
        } else
    #endif // AVX
    #ifndef __amd64__
        if(unlikely(!usableSSE2)){
            /* Pure C. */
            memset(this->bitmapAddr, 0, this->bitmapSize);
        } else
    #endif
        {
            asm volatile(
                "pxor %%xmm0, %%xmm0;"
                ".align 16;"
                "SSE2_LOOP:" /* memset 128bytes per LOOP. */
                    "movntdq %%xmm0, -0x80(%1, %0, 1);"
                    "movntdq %%xmm0, -0x70(%1, %0, 1);"
                    "movntdq %%xmm0, -0x60(%1, %0, 1);"
                    "movntdq %%xmm0, -0x50(%1, %0, 1);"
                    "movntdq %%xmm0, -0x40(%1, %0, 1);"
                    "movntdq %%xmm0, -0x30(%1, %0, 1);"
                    "movntdq %%xmm0, -0x20(%1, %0, 1);"
                    "movntdq %%xmm0, -0x10(%1, %0, 1);"
                #ifdef __amd64__
                    "subq $0x80, %0;"
                #else // __i386__
                    "subl $0x80, %0;"
                #endif
                "jnz SSE2_LOOP;"
                :
                : "r" (this->bitmapSize), "r" (this->bitmapAddr)
                : "cc", "%xmm0");
        }
    
    /* Reset advise. */
    madvise(this->bitmapAddr, this->bitmapSize, MADV_RANDOM);
}

