/*
 * GCDialog.java
 * Created on 2011/10/13
 *
 * Copyright (C) 2011-2013 Nippon Telegraph and Telephone Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */

package jp.co.ntt.oss.heapstats.gui;

import java.awt.Color;
import java.awt.Component;
import java.io.IOException;
import java.text.NumberFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JDialog;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JTable;
import javax.swing.SwingWorker;
import javax.swing.table.DefaultTableModel;

import jp.co.ntt.oss.heapstats.container.SnapShotHeader;
import jp.co.ntt.oss.heapstats.gui.snapshot.SnapShotReader;

import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.axis.CategoryAxis;
import org.jfree.chart.axis.CategoryLabelPositions;
import org.jfree.chart.axis.NumberAxis;
import org.jfree.chart.plot.CategoryPlot;
import org.jfree.chart.plot.DatasetRenderingOrder;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.chart.renderer.category.LineAndShapeRenderer;
import org.jfree.chart.renderer.category.StackedAreaRenderer;
import org.jfree.chart.renderer.category.AreaRenderer;
import org.jfree.data.category.DefaultCategoryDataset;

/**
 * extends {@link JDialog}<br>
 * <br>
 * Displays statistics for garbage collection.
 */
public class GCDialog extends JDialog {
    /** serialVersionUID. */
    private static final long serialVersionUID = 3024831843243414334L;

    /** Angle of the axis labels. */
    private static final double ANGLE = 6.0;

    /** Statistics table Number of rows. */
    private static final int SUMMARY_ROW = 5;

    /** Statistics table Number of columns. */
    private static final int SUMMARY_COL = 2;

    /** Mainframe heapstats. */
    private GUIMain guimain;

    /** Main Component. */
    private Box comp = Box.createVerticalBox();

    /** Minimum GC Time. */
    private long min = 0;

    /** Maximum GC Time. */
    private long max = 0;

    /** Panel of chart statistics. */
    private ChartPanel gcPanel;

    /** Panel of metaspace chart. */
    private ChartPanel metaspacePanel;

    /**
     * Create the 'GC Summary Dialog'.
     *
     * @param owner the owner Dialog from which the dialog is displayed or null
     *        if this dialog has no owner
     */
    public GCDialog(final JFrame owner) {
        super(owner);
        guimain = (GUIMain) owner;

        setTitle(GuiCommon.getResources()
                .getString(GuiCommon.TAB_TITLE_SUMMARY));
        setSize(GuiCommon.DEFAULT_DIALOG_WIDTH, GuiCommon.DEFAULT_DIALOG_HEIGHT);
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE);
        setModal(false);

        comp.setBorder(BorderFactory.createEmptyBorder(GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE, GuiCommon.BORDER_SIZE,
                GuiCommon.BORDER_SIZE));
        setLocationByPlatform(true);
    }

    /**
     * Draw a pie chart in order of the heap usage of the snapshot. In addition,
     * the table shows the heap usage for each object.
     *
     * @param diffOffsets file position and SnapShotReader
     */
    public final void draw(final List<SnapShotReader> diffOffsets) {
        clear();

        GcWorker worker = new GcWorker(this, diffOffsets);
        worker.execute();
    }

    /**
     * Clear the Summary data.
     */
    public final void clear() {
        for (Component obj : comp.getComponents()) {
            comp.remove(obj);
        }
        max = 0;
        min = 0;
        comp.setVisible(false);
    }

    /**
     * Return the component.
     *
     * @return Return the component
     */
    public final Box getComp() {
        return comp;
    }

    /**
     * Return the panel of chart statistics.
     *
     * @return Return the panel of chart statistics
     */
    public final ChartPanel getGcPanel() {
        return gcPanel;
    }

    /**
     * Sets the dialog component.
     *
     * @param value the Main Component
     */
    public final void setComp(final Box value) {
        comp = value;
        add(comp);
    }

    /**
     * extends {@link SwingWorker}<br>
     * <br>
     * SwingWorker thread drawing done in the time-consuming process chart.
     */
    class GcWorker extends SwingWorker<Object, Object> {

        /** file position and SnapShotReader. */
        private List<SnapShotReader> diffOffsets;

        /** the owner dialog. */
        private GCDialog dialog;

        /**
         * Create a GcWorker.
         *
         * @param owner determines the Dialog in which the Dialog is displayed;
         *        if null, or if the parentComponent has no Dialog, a default
         *        Dialog is used
         * @param offsets file position and SnapShotReader
         */
        public GcWorker(final GCDialog owner, final List<SnapShotReader> offsets) {
            diffOffsets = offsets;
            dialog = owner;
        }

        @Override
        protected String doInBackground() {
            try {
                ExpansionContractionPanel chart = new ExpansionContractionPanel(
                        GuiCommon.getResources().getString(
                                GuiCommon.TITLE_STATISTICS));
                ExpansionContractionPanel metaChart =
                                new ExpansionContractionPanel(
                                         GuiCommon.getResources().getString(
                                                     GuiCommon.TITLE_METASPACE));
                ExpansionContractionPanel summary = new ExpansionContractionPanel(
                        GuiCommon.getResources().getString(
                                GuiCommon.TITLE_SUMMARY));

                comp.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
                comp.add(chart);
                comp.revalidate();
                comp.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
                comp.add(metaChart);
                comp.revalidate();
                comp.add(Box.createVerticalStrut(GuiCommon.BORDER_SIZE));
                comp.add(summary);
                comp.revalidate();

                chart.setComponent(drawStatistictChart(), true);
                metaChart.setComponent(drawMetaspaceChart(), true);

                // Summary table
                DefaultTableModel model = new DefaultTableModel(SUMMARY_ROW,
                        SUMMARY_COL);
                JTable table = new JTable(model);
                summary.setComponent(table, true);

                model.setRowCount(0);
                Object[] row = new Object[2];

                // SnapShot Count
                NumberFormat nbf = NumberFormat.getIntegerInstance();
                row[0] = GuiCommon.getResources()
                        .getString(GuiCommon.SNAPCOUNT);
                row[1] = nbf.format(diffOffsets.size());
                model.addRow(row);

                // Garbage collection Count
                row[0] = GuiCommon.getResources().getString(GuiCommon.GCCOUNT);
                // List<Object> data = diffOffsets.get(diffOffsets.size() - 1);
                SnapShotReader reder = diffOffsets.get(diffOffsets.size() - 1);
                SnapShotHeader header = reder.getHeader();
                StringBuilder buf = new StringBuilder();
                buf.append(nbf.format(header.getFullCount()
                        + header.getYngCount()));
                buf.append("(Full:");
                buf.append(nbf.format(header.getFullCount()));
                buf.append(" , Young:");
                buf.append(nbf.format(header.getYngCount()));
                buf.append(")");
                row[1] = buf.toString();
                model.addRow(row);

                // Latest Heap Usage
                row[0] = GuiCommon.getResources().getString(
                        GuiCommon.LATEST_HEAP);
                row[1] = nbf.format((header.getNewHeap() + header.getOldHeap())
                        / GuiCommon.BINARY_PREFIX)
                        + "  KB";
                model.addRow(row);

                // Latest Metaspace Usage
                row[0] = GuiCommon.getResources().getString(
                                                   GuiCommon.LATEST_METASPACE);
                row[1] = nbf.format((header.getMetaspaceUsage())
                        / GuiCommon.BINARY_PREFIX)
                        + "  KB";
                model.addRow(row);

                // min gc time
                row[0] = GuiCommon.getResources().getString(
                        GuiCommon.MIN_GCTIME);
                row[1] = nbf.format(min) + "  msec";
                model.addRow(row);

                // max gc time
                row[0] = GuiCommon.getResources().getString(
                        GuiCommon.MAX_GCTIME);
                row[1] = nbf.format(max) + "  msec";
                model.addRow(row);

                // If snapshot does not have metaspace data, metaspace panel
                // is closed by default.
                if(header.getMetaspaceCapacity() == 0){
                  metaChart.mousePressed(null);
                }

                if (dialog.isVisible()) {
                    comp.setVisible(true);
                }
            } catch (IOException e) {
                GuiCommon.showMessage(dialog.getParent(), GuiCommon.BASEE12,
                        JOptionPane.ERROR_MESSAGE);
                e.printStackTrace();
            }

            return null;
        }

        /**
         * The chart displays the statistics for Garbage collection.
         *
         * @return Return the ChartPanel
         * @throws IOException If an I/O error has occurred.
         */
        private ChartPanel drawStatistictChart() throws IOException {

            DefaultCategoryDataset stackDataSet = new DefaultCategoryDataset();
            DefaultCategoryDataset usageDataSet = new DefaultCategoryDataset();
            DefaultCategoryDataset gctimeDataSet = new DefaultCategoryDataset();

            StringBuilder labelStr = new StringBuilder();

            SimpleDateFormat sdf = new SimpleDateFormat(
                    GuiCommon.CHART_DATE_FORMAT);

            List<String> removeLabel = new ArrayList<String>();

            int i = 0;
            int j = diffOffsets.size() / 10;
            for (SnapShotReader reader : diffOffsets) {

                SnapShotHeader header = reader.getHeader();

                String before = labelStr.toString();
                labelStr.delete(0, labelStr.length());
                labelStr.append(sdf.format(header.getSnapShotDate()));

                // set total heap size
                stackDataSet.addValue((double) header.getTotalSize()
                        / GuiCommon.BINARY_PREFIX, GuiCommon.getResources()
                        .getString(GuiCommon.TOTAL_HEAP_SIZE), labelStr
                        .toString());
                // new heap usage
                usageDataSet.addValue((double) header.getNewHeap()
                        / GuiCommon.BINARY_PREFIX, GuiCommon.getResources()
                        .getString(GuiCommon.NEW_HEAP), labelStr.toString());
                // old heap usage
                usageDataSet.addValue((double) header.getOldHeap()
                        / GuiCommon.BINARY_PREFIX, GuiCommon.getResources()
                        .getString(GuiCommon.OLD_HEAP), labelStr.toString());
                // gc time
                gctimeDataSet.addValue(header.getGcTime(), GuiCommon
                        .getResources().getString(GuiCommon.GCTIME), labelStr
                        .toString());

                // min gc time
                if (min == 0
                        || (min > header.getGcTime() && header.getGcTime() > 0)) {
                    min = header.getGcTime();
                }

                // max gc time
                if (max < header.getGcTime()) {
                    max = header.getGcTime();
                }

                if (!before.equals(labelStr.toString()) && (j > 0)
                        && (i++ % j != 0)) {
                    removeLabel.add(labelStr.toString());
                }
            }

            JFreeChart chart = ChartFactory.createStackedAreaChart("", "",
                    "Usage(KB)", stackDataSet, PlotOrientation.VERTICAL, true,
                    true, true);
            chart.setBackgroundPaint(Color.WHITE);

            CategoryPlot plot = chart.getCategoryPlot();
            plot.setBackgroundPaint(new Color(GuiCommon.getBgColor()));

            // Sets the DataSet heap usage(new + old).
            plot.setDataset(1, usageDataSet);

            // Sets the DataSet gc processing time.
            plot.setDataset(2, gctimeDataSet);

            // Sets vertical axis of the processing time value gc.
            plot.mapDatasetToRangeAxis(2, 2);

            NumberAxis axis = new NumberAxis("Time(msec)");
            axis.setStandardTickUnits(NumberAxis.createIntegerTickUnits());
            // gc processing time : Integer value
            plot.setRangeAxis(2, axis);

            StackedAreaRenderer renderer = new StackedAreaRenderer();

            LineAndShapeRenderer gcrnderer = new LineAndShapeRenderer();
            gcrnderer.setSeriesShapesVisible(0, false);

            plot.setRenderer(1, renderer);
            plot.setRenderer(2, gcrnderer);
            plot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);

            CategoryAxis ca = plot.getDomainAxis();
            ca.setCategoryLabelPositions(CategoryLabelPositions
                    .createUpRotationLabelPositions(Math.PI / ANGLE));

            // clear the unwanted text.
            for (String label : removeLabel) {
                ca.setTickLabelPaint(label, new Color(0, 0, 0, 0));
            }

            gcPanel = new ChartPanel(chart);

            gcPanel.addChartMouseListener(guimain);

            return gcPanel;
        }

        /**
         * The chart displays the statistics for Metaspace.
         *
         * @return Return the ChartPanel
         * @throws IOException If an I/O error has occurred.
         */
        private ChartPanel drawMetaspaceChart() throws IOException {
          DefaultCategoryDataset usageDataSet = new DefaultCategoryDataset();
          DefaultCategoryDataset capacityDataSet = new DefaultCategoryDataset();
          StringBuilder labelStr = new StringBuilder();
          SimpleDateFormat sdf = new SimpleDateFormat(
                                        GuiCommon.CHART_DATE_FORMAT);

          List<String> removeLabel = new ArrayList<String>();

          int i = 0;
          int j = diffOffsets.size() / 10;
          for(SnapShotReader reader : diffOffsets){
            SnapShotHeader header = reader.getHeader();

            String before = labelStr.toString();
            labelStr.delete(0, labelStr.length());
            labelStr.append(sdf.format(header.getSnapShotDate()));

            // set metaspace usage
            usageDataSet.addValue((double) header.getMetaspaceUsage()
                   / GuiCommon.BINARY_PREFIX, GuiCommon.getResources()
                     .getString(GuiCommon.METASPACE_USAGE), labelStr.toString());
            // set metaspace capacity
            capacityDataSet.addValue((double) header.getMetaspaceCapacity()
                  / GuiCommon.BINARY_PREFIX, GuiCommon.getResources()
                  .getString(GuiCommon.METASPACE_CAPACITY), labelStr.toString());

            if(!before.equals(labelStr.toString()) && (j > 0) && (i++ % j != 0)){
              removeLabel.add(labelStr.toString());
            }
          }

          JFreeChart chart = ChartFactory.createStackedAreaChart("", "",
                    "Usage(KB)", capacityDataSet, PlotOrientation.VERTICAL,
                                                              true, true, true);
          chart.setBackgroundPaint(Color.WHITE);

          CategoryPlot plot = chart.getCategoryPlot();
          plot.setBackgroundPaint(new Color(GuiCommon.getBgColor()));

          // Set the metaspace usage dataset
          plot.setDataset(1, usageDataSet);
          StackedAreaRenderer renderer = new StackedAreaRenderer();
          plot.setRenderer(1, renderer);
          plot.setDatasetRenderingOrder(DatasetRenderingOrder.FORWARD);

          CategoryAxis ca = plot.getDomainAxis();
          ca.setCategoryLabelPositions(CategoryLabelPositions
                    .createUpRotationLabelPositions(Math.PI / ANGLE));

          // clear the unwanted text.
          for(String label : removeLabel){
            ca.setTickLabelPaint(label, new Color(0, 0, 0, 0));
          }

          metaspacePanel = new ChartPanel(chart);

          metaspacePanel.addChartMouseListener(guimain);

          return metaspacePanel;
        }

    }
}
